from datetime import datetime
from django.db import models
from django.contrib.auth.models import (
    BaseUserManager,
    AbstractBaseUser,
    PermissionsMixin,
)
from django.core.exceptions import ValidationError

from web3 import Web3


def validate_ethereum_address(value):
    if not Web3.isChecksumAddress(value):
        raise ValidationError


class WalletManager(BaseUserManager):
    def create_user(self, ethereum_address: str):
        """
        Creates and saves a User with the given ethereum address.
        """
        if not ethereum_address:
            raise ValueError("Users must have an ethereum address")

        wallet = self.model()
        wallet.ethereum_address = ethereum_address

        wallet.save(using=self._db)
        return wallet

    def create_superuser(self, ethereum_address: str):
        """
        Creates and saves a superuser with the given ethereum address.
        """
        user = self.create_user(
            ethereum_address=ethereum_address,
        )
        user.set_unusable_password()
        user.is_admin = True
        user.is_superuser = True
        user.save(using=self._db)
        return user


class Wallet(AbstractBaseUser, PermissionsMixin):
    # EIP-55 compliant: https://eips.ethereum.org/EIPS/eip-55
    ethereum_address = models.CharField(
        unique=True,
        primary_key=True,
        max_length=42,
        validators=[validate_ethereum_address],
    )
    ens_name = models.CharField(max_length=255, blank=True, null=True)
    ens_avatar = models.CharField(max_length=255, blank=True, null=True)
    created = models.DateTimeField("datetime created", auto_now_add=True)
    is_active = models.BooleanField(default=True)
    is_admin = models.BooleanField(default=False)

    USERNAME_FIELD = "ethereum_address"

    objects = WalletManager()

    def __str__(self):
        return self.ethereum_address

    @property
    def is_staff(self):
        "Is the user a member of staff?"
        # Simplest possible answer: All admins are staff
        return self.is_admin


class Nonce(models.Model):
    value = models.CharField(max_length=24, primary_key=True)
    expiration = models.DateTimeField()

    def __str__(self):
        return self.value



from django import forms
from django.contrib import admin
from django.contrib.auth.models import Group
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.forms import ReadOnlyPasswordHashField
from django.core.exceptions import ValidationError

from .models import Wallet


class WalletCreationForm(forms.ModelForm):
    class Meta:
        model = Wallet
        fields = ("ethereum_address",)

    def save(self, commit=True):
        # Save the provided password in hashed format
        user = super().save(commit=False)
        if commit:
            user.save()
        return user


class WalletChangeForm(forms.ModelForm):
    """A form for updating users. Includes all the fields on
    the user, but replaces the password field with admin's
    disabled password hash display field.
    """

    class Meta:
        model = Wallet
        fields = ("ethereum_address", "ens_name", "ens_avatar", "is_active", "is_admin")


class WalletAdmin(BaseUserAdmin):
    # The forms to add and change user instances
    form = WalletChangeForm
    add_form = WalletCreationForm

    list_display = (
        "ethereum_address",
        "ens_name",
        "ens_avatar",
        "is_active",
        "is_admin",
    )
    list_filter = (
        "is_active",
        "is_admin",
    )
    fieldsets = (
        (None, {"fields": ("ethereum_address",)}),
        ("Permissions", {"fields": ("is_admin",)}),
    )
    add_fieldsets = (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("ethereum_address",),
            },
        ),
    )
    search_fields = (
        "ethereum_address",
        "ens_name",
    )
    ordering = (
        "ethereum_address",
        "ens_name",
    )
    filter_horizontal = ()


# Now register the new UserAdmin...
admin.site.register(Wallet, WalletAdmin)



from datetime import datetime, timedelta
import secrets
import pytz

from django.shortcuts import redirect

from django.conf import settings
from django.http import JsonResponse
from django.contrib.auth import authenticate
from django.contrib.auth import login as auth_login
from django.contrib.auth import logout as auth_logout
from django.views.decorators.http import require_http_methods

from ratelimit.decorators import ratelimit

from .custom_groups.group_manager import GroupManager
from .models import Nonce, Wallet


@ratelimit(key='ip', rate='5/m')
@require_http_methods(["POST"])
def login(request):
    wallet = authenticate(request)
    if wallet is not None:
        if wallet.is_active:
            auth_login(request, wallet)
            return JsonResponse({"success": True, "message": "Successful login."})
        else:
            return JsonResponse(
                {"success": False, "message": "Wallet disabled."}, status=401
            )
    return JsonResponse({"success": False, "message": "Invalid login."}, status=403)


def _check_groups(wallet: Wallet):
    for group in settings.CUSTOM_GROUPS:
        name: str = group[0]
        manager: GroupManager = group[1]

        if manager.is_member(ethereum_address=wallet.ethereum_address):
            wallet.groups.add(name)
        else:
            wallet.groups.remove(name)


@ratelimit(key='ip', rate='5/m')
@require_http_methods(["POST"])
def logout(request):
    auth_logout(request)
    return redirect("/")


@ratelimit(key='ip', rate='5/m')
@require_http_methods(["GET"])
def nonce(request):
    now = datetime.now(tz=pytz.UTC)

    _scrub_nonce()
    n = Nonce(value=secrets.token_hex(12), expiration=now + timedelta(hours=12))
    n.save()

    return JsonResponse({"nonce": n.value})


def _scrub_nonce():
    # Delete all expired nonce's
    for n in Nonce.objects.filter(expiration__lte=datetime.now(tz=pytz.UTC)):
        n.delete()



# Generated by Django 4.0.5 on 2022-06-10 18:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('siwe_auth', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='wallet',
            name='created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='datetime created'),
        ),
    ]



# Generated by Django 4.0.4 on 2022-06-01 09:45

from django.db import migrations, models
import siwe_auth.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='Nonce',
            fields=[
                ('value', models.CharField(max_length=24, primary_key=True, serialize=False)),
                ('expiration', models.DateTimeField()),
            ],
        ),
        migrations.CreateModel(
            name='Wallet',
            fields=[
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('ethereum_address', models.CharField(max_length=42, primary_key=True, serialize=False, unique=True, validators=[siwe_auth.models.validate_ethereum_address])),
                ('ens_name', models.CharField(blank=True, max_length=255, null=True)),
                ('ens_avatar', models.CharField(blank=True, max_length=255, null=True)),
                ('created', models.DateTimeField(verbose_name='datetime created')),
                ('is_active', models.BooleanField(default=True)),
                ('is_admin', models.BooleanField(default=False)),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.permission', verbose_name='user permissions')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from abc import abstractmethod
import logging
from typing import Callable

from web3 import Web3, HTTPProvider

from siwe_auth.custom_groups.group_manager import GroupManager


class ERC721Manager(GroupManager):
    contract: str
    abi = [
        {
            "constant": False,
            "inputs": [{"name": "_owner", "type": "address"}],
            "name": "balanceOf",
            "outputs": [{"name": "", "type": "uint256"}],
            "payable": False,
            "type": "function",
        }
    ]

    def __init__(self, config: dict):
        if "contract" not in config:
            raise ValueError("ERC721 Manager config is missing contract attribute.")
        self.contract = config["contract"]

    def _is_member(
        self,
        ethereum_address: str,
        provider: HTTPProvider,
        expression: Callable[[str], bool],
    ):
        w3 = Web3(provider)
        contract = w3.eth.contract(
            address=Web3.toChecksumAddress(self.contract.lower()), abi=self.abi
        )
        balance = contract.functions.balanceOf(
            _owner=Web3.toChecksumAddress(ethereum_address.lower())
        ).call()
        return expression(balance)

    @abstractmethod
    def is_member(self, wallet: object, provider: HTTPProvider) -> bool:
        pass


class ERC721OwnerManager(ERC721Manager):
    def is_member(self, wallet: object, provider: HTTPProvider) -> bool:
        if not self._valid_wallet(wallet=wallet):
            return False
        try:
            return self._is_member(
                ethereum_address=wallet.ethereum_address,
                provider=provider,
                expression=lambda x: x > 0,
            )
        except ValueError:
            logging.error(
                f"Unable to verify membership of invalid address: {wallet.ethereum_address}"
            )
        return False



from django.urls import re_path

from . import views

app_name = "siwe_auth"
urlpatterns = [
    re_path(r"^login/?$", views.login, name="login"),
    re_path(r"^logout/?$", views.logout, name="logout"),
    re_path(r"^nonce/?$", views.nonce, name="nonce"),
]



import json
import datetime
import logging
import re
from typing import Optional

import pytz

from django.conf import settings
from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.models import User, Group
from web3 import Web3, HTTPProvider
from web3.middleware import geth_poa_middleware
from ens import ENS

from siwe.siwe import (
    SiweMessage,
    ValidationError,
    ExpiredMessage,
    MalformedSession,
    InvalidSignature,
)

from .custom_groups.group_manager import GroupManager
from .models import Wallet, Nonce


def _nonce_is_valid(nonce: str) -> bool:
    """
    Check if given nonce exists and has not yet expired.
    :param nonce: The nonce string to validate.
    :return: True if valid else False.
    """
    n = Nonce.objects.get(value=nonce)
    is_valid = False
    if n is not None and n.expiration > datetime.datetime.now(tz=pytz.UTC):
        is_valid = True
    n.delete()
    return is_valid


class SiweBackend(BaseBackend):
    """
    Authenticate an Ethereum address as per Sing-In with Ethereum (EIP-4361).
    """

    def authenticate(self, request, signature: str = None, siwe_message: SiweMessage = None):
        body = json.loads(request.body)

        if siwe_message is None:
            siwe_message = SiweMessage(
                message={
                    re.sub(r"(?<!^)(?=[A-Z])", "_", k).lower(): v
                    for k, v in body["message"].items()
                }
            )
            signature = body["signature"]

        # Validate signature
        w3 = Web3(HTTPProvider(settings.PROVIDER))
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)

        try:
            siwe_message.validate(signature=signature, provider=w3)
        except ValidationError:
            logging.info("Authentication attempt rejected due to invalid message.")
            return None
        except ExpiredMessage:
            logging.info("Authentication attempt rejected due to expired message.")
            return None
        except MalformedSession as e:
            logging.info(
                f"Authentication attempt rejected due to missing fields: {', '.join(e.missing_fields)}"
            )
            return None
        except InvalidSignature:
            logging.info("Authentication attempt rejected due to invalid signature.")
            return None

        # Validate nonce
        if not _nonce_is_valid(siwe_message.nonce):
            return None

        # Pull ENS data
        if getattr(settings, "CREATE_ENS_PROFILE_ON_AUTHN", True):
            ens_profile = ENSProfile(ethereum_address=siwe_message.address, w3=w3)
        else:
            ens_profile = ENSProfile.__new__(ENSProfile) # blank ENSProfile, skipping __init__ constructor


        # Message and nonce has been validated. Authentication complete. Continue with authorization/other.
        now = datetime.datetime.now(tz=pytz.UTC)
        try:
            wallet = Wallet.objects.get(ethereum_address=siwe_message.address)
            wallet.last_login = now
            wallet.ens_name = ens_profile.name
            wallet.save()
            logging.debug(f"Found wallet for address {siwe_message.address}")
        except Wallet.DoesNotExist:
            wallet = Wallet(
                ethereum_address=Web3.toChecksumAddress(siwe_message.address),
                ens_name=ens_profile.name,
                ens_avatar=ens_profile.avatar,
                last_login=now,
                password=None,
            )
            wallet.set_unusable_password()
            wallet.save()
            logging.debug(
                f"Could not find wallet for address {siwe_message.address}. Creating new wallet object."
            )

        # Group settings
        if getattr(settings, "CREATE_GROUPS_ON_AUTHN", False):
            for custom_group in settings.CUSTOM_GROUPS:
                group, created = Group.objects.get_or_create(name=custom_group[0])
                if created:
                    logging.info(f"Created group '{custom_group[0]}'.")
                group_manager: GroupManager = custom_group[1]
                if group_manager.is_member(
                    wallet=wallet,
                    provider=HTTPProvider(settings.PROVIDER),
                ):
                    logging.info(
                        f"Adding wallet '{wallet.ethereum_address}' to group '{custom_group[0]}'."
                    )
                    wallet.groups.add(group)

        return wallet

    def get_user(self, ethereum_address: str) -> Optional[Wallet]:
        """
        Get Wallet by ethereum address if exists.
        :param ethereum_address: Ethereum address of user.
        :return: Wallet object if exists or None
        """
        try:
            return Wallet.objects.get(pk=ethereum_address)
        except User.DoesNotExist:
            return None


class ENSProfile:
    """
    Container for ENS profile information including but not limited to primary name and avatar.
    """

    name: str = None
    avatar: str = None

    def __init__(self, ethereum_address: str, w3: Web3):
        # Temporary until https://github.com/ethereum/web3.py/pull/2286 is merged
        self.name = ENS.fromWeb3(w3).name(address=ethereum_address)
        resolver = ENS.fromWeb3(w3).resolver(normal_name=self.name)
        # if resolver:
        #     self.avatar = resolver.caller.text(normal_name_to_hash(self.name), 'avatar')



from abc import ABC, abstractmethod
from web3 import HTTPProvider


class GroupManager(ABC):
    @abstractmethod
    def __init__(self, config: dict):
        """
        GroupManager initialization function.
        :param config: Dictionary for passing in any dependencies such as contract address or list of valid addresses.
        """
        pass

    @abstractmethod
    def is_member(self, wallet: object, provider: HTTPProvider) -> bool:
        """
        Membership function to identify if a given ethereum address is part of this class' group.
        :param provider: Web3 provider to use for membership check.
        :param wallet: Object with ethereum_address attribute to check membership of.
        :return: True if address is a member else False
        """
        pass

    def _valid_wallet(self, wallet: object):
        return wallet.__getattribute__('ethereum_address') is not None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['siwe_auth',
 'siwe_auth.custom_groups',
 'siwe_auth.migrations',
 'siwe_auth.utils']

package_data = \
{'': ['*'], 'siwe_auth': ['static/siwe_auth/js/*', 'templates/siwe_auth/*']}

install_requires = \
['Django>=4.0.1,<5.0.0',
 'django-ratelimit>=3.0.1,<4.0.0',
 'pytz>=2021.3,<2022.0',
 'siwe>=1.0.0,<2.0.0',
 'web3>=5.26.0,<6.0.0']

setup_kwargs = {
    'name': 'django-siwe-auth',
    'version': '0.3.0',
    'description': 'Custom Django authentication backend using Sign-In with Ethereum (EIP-4361) with a custom wallet user model.',
    'long_description': '<div id="top"></div>\n\n<!-- PROJECT SHIELDS -->\n[![Contributors][contributors-shield]][contributors-url]\n[![Forks][forks-shield]][forks-url]\n[![Stargazers][stars-shield]][stars-url]\n[![Issues][issues-shield]][issues-url]\n[![MIT License][license-shield]][license-url]\n\n<!-- PROJECT LOGO -->\n<br />\n<div align="center">\n  <a href="https://github.com/payton/django-siwe-auth">\n    <img src="images/logo.png" alt="Logo" width="400" height="auto">\n  </a>\n<h3 align="center">Django Sign-In with Ethereum Auth</h3>\n  <p align="center">\n    A complete Django authentication system overhaul for Web3. Treat Ethereum wallets as first-class citizens with authentication via Sign-In with Ethereum (EIP-4361) and authorization via on-chain attributes such as NFT (ERC721/ERC1155) and ERC20 token ownership. \n    <br />\n    <h4><b><u>DISCLAIMER: django-siwe-auth is still in early development (it will change often!)</u></b></h4>\n    <a href="https://github.com/payton/django-siwe-auth"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    <a href="https://github.com/payton/django-siwe-auth">Live Demo</a>\n    ·\n    <a href="https://github.com/payton/django-siwe-auth/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/payton/django-siwe-auth/issues">Request Feature</a>\n  </p>\n</div>\n\n<!-- TABLE OF CONTENTS -->\n<details>\n  <summary>Table of Contents</summary>\n  <ol>\n    <li>\n      <a href="#about-the-project">About The Project</a>\n    </li>\n    <li>\n      <a href="#getting-started">Getting Started</a>\n    </li>\n    <li>\n        <a href="#usage">Usage</a>\n        <ul>\n            <li><a href="#useful-resources">Useful Resources</a></li>\n            <li><a href="#install">Install</a></li>\n            <li><a href="#rest-api">REST API</a></li>\n            <li><a href="#configuring-your-project">Configuring Your Project</a></li>\n            <li><a href="#group-plugins">Group Plugins</a></li>\n        </ul>\n    </li>\n    <li><a href="#contributing">Scenarios</a></li>\n    <li><a href="#contributing">Contributing</a></li>\n    <li><a href="#license">License</a></li>\n    <li><a href="#contact">Contact</a></li>\n    <li><a href="#acknowledgments">Acknowledgments</a></li>\n    <li><a href="#disclaimer">Disclaimer</a></li>\n  </ol>\n</details>\n\n<!-- ABOUT THE PROJECT -->\n## About The Project\n\nThis app provides four key features for new or existing Django projects.\n\n<ol>\n<li><b>Authentication Backend</b></li>\n<ul>\n<li>Authenticate incoming requests via the Sign-In with Ethereum (<a href="https://eips.ethereum.org/EIPS/eip-4361">EIP-4361</a>) standard.</li>\n</ul>\n<li><b>User Model</b></li>\n<ul>\n<li>Replace original user model with wallet model that is dependent on an Ethereum address.</li>\n<li>Removes the liability of storing user credentials in favor of managing transient session data (<a href="https://docs.djangoproject.com/en/4.0/topics/http/sessions/">handled by Django out of the box</a>).</li>\n<li><a href="https://ens.domains/">ENS</a> profile information is pulled by default.</li>\n</ul>\n<li><b>Groups</b></li>\n<ul>\n<li>Create custom user groups based on on-chain (or off-chain!) attributes.</li>\n<li>For example, we can easily define a group of <a href="https://boredapeyachtclub.com/#/">BAYC</a> owners by validating NFT ownership and subsequently serve them special content that non-owners don\'t have access to.</li>\n<li>This extension builds off of Django\'s well-defined <a href="https://docs.djangoproject.com/en/4.0/topics/auth/default/#permissions-and-authorization">authorization system</a>.</li>\n</ul>\n</ol>\n\nBelow is the included example application that authenticates with an Ethereum address and utilizes on-chain attributes to authorize access to various notepads.\n<div align="center">\n   <img src="images/example.gif" alt="Example Project" width="800" height="auto">\n</div>\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n<!-- GETTING STARTED -->\n## Getting Started\n\nWith the included example applications, you can test out Sign-In with Ethereum along with using and creating custom groups. To get an example application up and running follow these steps.\n\n### Prerequisites\n\nRequirements for developing and running examples apps:\n* [poetry](https://python-poetry.org/docs/)\n* [npm](https://www.npmjs.com/)\n\n### Demo Application\n\n1. Install NPM dependencies\n   ```sh\n   npm --prefix examples/notepad/frontend install examples/notepad/frontend\n   ```\n2. Build frontend\n   ```sh\n   npm run --prefix examples/notepad/frontend build\n   ```\n3. Install Python dependencies\n   ```sh\n   poetry install\n   ```\n4. Set Web3 provider environment variable\n   ```sh\n   export SIWE_AUTH_PROVIDER="https://mainnet.infura.io/v3/..." # Any provider works\n   ```\n5. Django migrations\n   ```sh\n   # Create Django migrations\n   poetry run examples/notepad/manage.py makemigrations\n   # Apply Django migrations\n   poetry run examples/notepad/manage.py migrate\n   ```\n6. Run server\n   ```sh\n   poetry run examples/notepad/manage.py runserver\n   ```\n7. Test application at `https://localhost:8000`\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n### Rebuild Demo JavaScript\nIf you make an update to the frontend directory or siwe submodule, rebuild bundle.js:\n```sh\ncd examples/notepad/frontend \\\nnpm install \\\nnpm run build\n```\n\n\n<!-- USAGE EXAMPLES -->\n## Usage\n\nThis project is highly configurable and modular allowing for applications to be as simple or complex as required. Authentication and authorization can be completely replaced or supplemented depending upon the use case. See scenario docs for detailed tutorials.\n\n### Useful Resources\n* [User authentication in Django](https://docs.djangoproject.com/en/4.0/topics/auth/)\n* [Customizing authentication in Django](https://docs.djangoproject.com/en/4.0/topics/auth/customizing/)\n* [How to Switch to a Custom Django User Model Mid-Project](https://www.caktusgroup.com/blog/2019/04/26/how-switch-custom-django-user-model-mid-project/)\n\n### Install\n\n#### Prerequisites\n\nRequirements for using `django-siwe-auth` in a Django application:\n* [pip](https://pip.pypa.io/en/stable/installation/)\n* [django](https://docs.djangoproject.com/en/4.0/intro/install/)\n* An existing [Django project](https://docs.djangoproject.com/en/4.0/intro/tutorial01/)\n\n1. Install package\n   ```sh\n   pip install django-siwe-auth\n   ```\n2. Add `siwe_auth.apps.SiweAuthConfig` to INSTALLED_APPS in your project\'s `settings.py`\n   ```python\n   INSTALLED_APPS = [\n    ...\n    "siwe_auth.apps.SiweAuthConfig", # Adds django-siwe-auth\n    "django.contrib.admin",\n    "django.contrib.auth",\n    "django.contrib.contenttypes",\n    "django.contrib.sessions",\n    "django.contrib.messages",\n    "django.contrib.staticfiles",\n   ]\n   ```\n3. Add app-specific definitions to your project\'s `settings.py` (same files as previous step). For more information, see the <a href="#configuring-your-project">Configuring Your Project</a> section\n   ```python\n   from siwe_auth.custom_groups.erc721 import ERC721OwnerManager\n   ...\n   # Django Sign-In with Ethereum Auth Settings\n   AUTH_USER_MODEL = "siwe_auth.Wallet"\n   AUTHENTICATION_BACKENDS = ["siwe_auth.backend.SiweBackend"]\n   LOGIN_URL = "/"\n   SESSION_COOKIE_AGE = 3 * 60 * 60 \n   CREATE_GROUPS_ON_AUTHN = False # defaults to False\n   CREATE_ENS_PROFILE_ON_AUTHN = True # defaults to True\n   CUSTOM_GROUPS = [\n       (\'ens_owners\', ERC721OwnerManager(config={\'contract\': \'0x57f1887a8BF19b14fC0dF6Fd9B2acc9Af147eA85\'})),\n       (\'bayc_owners\', ERC721OwnerManager(config={\'contract\': \'0xBC4CA0EdA7647A8aB7C2061c2E118A18a936f13D\'})),\n       (\'shib_owners\', ERC20OwnerManager(config={\'contract\': \'0x95ad61b0a150d79219dcf64e1e6cc01f0b64c4ce\'})),\n   ]  # See "Group Plugins" section\n   PROVIDER = os.environ.get("SIWE_AUTH_PROVIDER", "https://mainnet.infura.io/v3/...")\n   ...\n   ```\n4. Add `api/auth/` endpoint to your project\'s `urls.py`\n   ```python\n   urlpatterns = [\n    path("admin/", admin.site.urls),\n    path("api/auth/", include("siwe_auth.urls")),\n    ...\n   ]\n   ```\n5. Run `manage.py makemigrations` and `manage.py migrate` to create custom user model\n6. Update frontend to interface with the <a href="#rest-api">REST API</a> as defined below. An example frontend is available in the notepad demo.\n7. Never handle a password ever again!\n\nFor more examples, please refer to the [Documentation](https://github.com/payton/django-siwe-auth)\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n\n### REST API\n\nProjects that use this app will interact via these endpoints for authentication. Tools for authorization are provided but not enforced via any endpoints.\n\n#### Login\n* Authenticate user based on signed message.\n\n<details>\n<summary>/api/auth/logout</summary>\n\n* **Method:**\n\n  `POST`\n\n* **URL Params**\n\n   None\n\n* **Data Params**\n\n  **Required:**\n  SIWE Message\n\n* **Success Response:**\n\n   * **Code:** 200 <br />\n     **Content:** `{"success": True, "message": "Successful login."}`\n\n* **Error Response:**\n\n   * **Code:** 401 UNAUTHORIZED <br />\n     **Content:** `{"success": False, "message": "Wallet disabled."}`\n\n  OR\n\n   * **Code:** 403 FORBIDDEN <br />\n     **Content:** `{"success": False, "message": "Invalid login."}`\n\n  OR\n\n    * **Code:** 429 TOO MANY REQUESTS <br />\n      **Content:** `{"message": "Too many requests. Slow down."}`\n\n* **Sample Call:**\n```javascript\nfetch(`/api/auth/login`, {\n    method: \'POST\',\n    body: JSON.stringify({ message }),\n    headers: { \'Content-Type\': \'application/json\' },\n    credentials: \'include\',\n}).then(async (res) => {\n    if (res.status === 200) {\n        // Success, continue with authenticated flow\n    } else {\n        // Failure, handle alternative path\n    }\n});\n```\n\n</details>\n\n#### Logout\n\n* Logout authenticated user.\n\n<details>\n<summary>/api/auth/logout</summary>\n\n* **Method:**\n\n  `POST`\n\n* **URL Params**\n\n  None\n\n* **Data Params**\n\n  None\n\n* **Success Response:**\n\n    * **Code:** 200 <br />\n      **Content:** `{"success": True, "message": "Successful login."}`\n\n* **Error Response:**\n\n    * **Code:** 401 UNAUTHORIZED <br />\n      **Content:** `{"success": False, "message": "Wallet disabled."}`\n\n  OR\n\n    * **Code:** 403 FORBIDDEN <br />\n      **Content:** `{"success": False, "message": "Invalid login."}`\n\n  OR\n\n    * **Code:** 429 TOO MANY REQUESTS <br />\n      **Content:** `{"message": "Too many requests. Slow down."}`\n\n* **Sample Call:**\n```javascript\nfetch(\'/api/authentication/logout\', {\n    method: \'POST\',\n    credentials: \'include\',\n}).then(() => {\n    // Success, continue with logout flow\n});\n```\n</details>\n\n#### Nonce\n\n* Get temporary nonce to include in EIP-4361 (SIWE) message.\n\n<details>\n<summary>/api/auth/nonce</summary>\n\n* **Method:**\n\n  `GET`\n\n* **URL Params**\n\n  None\n\n* **Data Params**\n\n  None\n\n* **Success Response:**\n\n    * **Code:** 200 <br />\n      **Content:** `{"success": True, "message": "Successful login."}`\n\n* **Error Response:**\n\n    * **Code:** 429 TOO MANY REQUESTS <br />\n      **Content:** `{"message": "Too many requests. Slow down."}`\n\n* **Sample Call:**\n```javascript\nfetch(\'/api/authentication/nonce\', { \n    credentials: \'include\' \n}).then((res) =>\n    // Success, continue with message signing flow\n);\n```\n</details>\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n### Configuring Your Project\n\n#### Relevant native Django settings\n\n```py\n# in settings.py\nAUTH_USER_MODEL = "siwe_auth.Wallet" # required for siwe as the default authentication\nAUTHENTICATION_BACKENDS = ["siwe_auth.backend.SiweBackend"] # required for siwe as the default authentication\nLOGIN_URL = "/" # optional, django\'s default is "/accounts/login/"\nSESSION_COOKIE_AGE = 3 * 60 * 60 # Age of cookie, in seconds. Optional, django\'s default is weeks.\n```\n\n#### django-siwe-auth specific settings\n```py\n# in settings.py\nCREATE_GROUPS_ON_AUTHN = True # optional, default is False\nCREATE_ENS_PROFILE_ON_AUTHN = True # optional, default is True\nCUSTOM_GROUPS = [] # optional, see "Adding a Group" below\n# Set environment variable SIWE_AUTH_PROVIDER for Web3.py\n# - Required if CREATE_GROUPS_ON_AUTHN or CREATE_ENS_PROFILE_ON_AUTHN are True. Optional otherwise.\n#   Any ethereum API key (infura or alchemy) will work.\nPROVIDER = os.environ.get("SIWE_AUTH_PROVIDER", "https://mainnet.infura.io/v3/...")\n```\n\n#### Override Admin Login Page\n\nIn order to login to the django admin page with siwe-auth, we need to override the login template.\n\n```python\n# in your app\'s admin.py, add the following line\nadmin.site.login_template = \'siwe_auth/login.html\'\n```\n\n\n<!-- USAGE EXAMPLES -->\n### Group Plugins\nA group plugin allows you to define your own group whose membership is defined by the output of a membership function. Some examples of what the membership function may check for include but are not limited to: NFT ownership, ERC-20 token ownership, ENS data, etc.\n\n#### GroupManager Class\nAll group plugins must implement the GroupManager class. This includes a function called `is_member` that is called to determine group membership when a user authenticates with the server.\n```python\nfrom abc import ABC, abstractmethod\n\nclass GroupManager(ABC):\n\n   @abstractmethod\n   def __init__(self, config: dict):\n      pass\n\n   @abstractmethod\n    def is_member(self, wallet: object) -> bool:\n        pass\n```\n\n#### Included Managers\n\n<ul>\n<li>ERC20Manager\n    <ul>\n        <li>ERC20OwnerManager</li>\n    </ul>\n</li>\n<li>ERC721Manager\n    <ul>\n        <li>ERC721OwnerManager</li>\n    </ul>\n</li>\n<li>ERC1155Manager\n    <ul>\n        <li>ERC1155OwnerManager</li>\n    </ul>\n</li>\n</ul>\n\n#### Adding a Group\nA custom group is defined by a tuple consisting of a name (string), MemberManager implementation.\n\nThe GroupManager\'s config can be used for anything, but some likely usecases are defining contract addresses or address include/exclude lists.\n\nIn your project\'s `settings.py`, append your group to the `CUSTOM_GROUPS` list added in the installation steps.\n\nSuppose we want to have only one group for `BAYC` (ERC721 NFT) owners and another group for `LPT` (ERC20 Token) owners. Our list would then look like:\n```python\nfrom siwe_auth.custom_groups.erc721 import ERC721OwnerManager\nfrom siwe_auth.custom_groups.erc20 import ERC20OwnerManager\n...\nCUSTOM_GROUPS=[\n   (\'bayc_owners\', ERC721OwnerManager(config={\'contract\': \'0xbc4ca0eda7647a8ab7c2061c2e118a18a936f13d\'})),\n   (\'lpt_owners\', ERC20OwnerManager(config={\'contract\': \'0x58b6a8a3302369daec383334672404ee733ab239\'})),\n]\n```\n\n\n\n## Scenarios (WIP)\n\nExamples:\n1. Start a new web app with wallet user model.\n2. Replace existing user model with wallet user model (non-trivial).\n3. Replace existing user model with weaker wallet user model that implements base model (less non-trivial but still non-trivial).\n4. Creating permissions around on-chain based groups.\n\n\n\n<!-- CONTRIBUTING -->\n## Contributing\n\nContributions are what make the open source community such an amazing place to learn, inspire, and create. Any contributions you make are **greatly appreciated**.\n\nIf you have a suggestion that would make this better, please fork the repo and create a pull request. You can also simply open an issue with the tag "enhancement".\nDon\'t forget to give the project a star! Thanks again!\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n<!-- LICENSE -->\n## License\n\nDistributed under the MIT License. See `LICENSE.txt` for more information.\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n<!-- CONTACT -->\n## Contact\n\npayton - [@paytongarland](https://twitter.com/paytongarland) - paytongarland.eth\n\nProject Link: [https://github.com/payton/django-siwe-auth](https://github.com/payton/django-siwe-auth)\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n\n\n<!-- ACKNOWLEDGMENTS -->\n## Acknowledgments\n\n* [Spruce](https://www.spruceid.com/) - creators of EIP-4361\n* [jsdelivr](https://www.jsdelivr.com/) - external static asset distribution in example app\n* [Best-README-Template](https://github.com/othneildrew/Best-README-Template)\n* [98.css](https://jdan.github.io/98.css/)\n* [API Documentation Format](https://gist.github.com/iros/3426278)\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n## Disclaimer\nThis django auth library has not yet undergone a formal security audit. We welcome continued feedback on the usability, architecture, and security of this implementation.\n\n<p align="right">(<a href="#top">back to top</a>)</p>\n\n<!-- MARKDOWN LINKS & IMAGES -->\n<!-- https://www.markdownguide.org/basic-syntax/#reference-style-links -->\n[contributors-shield]: https://img.shields.io/github/contributors/payton/django-siwe-auth.svg?style=for-the-badge\n[contributors-url]: https://github.com/payton/django-siwe-auth/graphs/contributors\n[forks-shield]: https://img.shields.io/github/forks/payton/django-siwe-auth.svg?style=for-the-badge\n[forks-url]: https://github.com/payton/django-siwe-auth/network/members\n[stars-shield]: https://img.shields.io/github/stars/payton/django-siwe-auth.svg?style=for-the-badge\n[stars-url]: https://github.com/payton/django-siwe-auth/stargazers\n[issues-shield]: https://img.shields.io/github/issues/payton/django-siwe-auth.svg?style=for-the-badge\n[issues-url]: https://github.com/payton/django-siwe-auth/issues\n[license-shield]: https://img.shields.io/github/license/payton/django-siwe-auth.svg?style=for-the-badge\n[license-url]: https://github.com/payton/django-siwe-auth/blob/main/LICENSE.txt\n[product-screenshot]: images/screenshot.png\n',
    'author': 'Payton Garland',
    'author_email': 'payton.r.g@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/payton/django-siwe-auth',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
